///////////////////////////////////////////////////////////////////////////////
// File: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/Tchar.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToString;

using JetByteTools::Socket::ConsecutiveSocketOperations;
using JetByteTools::Socket::PerSocketBufferPoolingLimit;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Constants
///////////////////////////////////////////////////////////////////////////////

static const _tstring s_serverArgument                   = _T("server");
static const _tstring s_portArgument                     = _T("port");
static const _tstring s_messageSizeArgument              = _T("messageSize");
static const _tstring s_messageRateArgument              = _T("messageRate");
static const _tstring s_maxBuffersPooledPerSocket        = _T("perSocketBufferPoolSize");
static const _tstring s_maxConsecutiveOperations         = _T("maxOps");
static const _tstring s_recvBufferSizeArgument           = _T("recvBuffer");
static const _tstring s_sendBufferSizeArgument           = _T("sendBuffer");
static const _tstring s_maxPendingWritesArgument         = _T("maxPendingWrites");
static const _tstring s_syncConnectSwitch                = _T("syncConnect");


///////////////////////////////////////////////////////////////////////////////
// CCommandLine
///////////////////////////////////////////////////////////////////////////////

CCommandLine::CCommandLine()
{
   m_parser.AddArgument(s_serverArgument);
   m_parser.AddArgument(s_portArgument);
   m_parser.AddArgument(s_messageSizeArgument);
   m_parser.AddArgument(s_messageRateArgument);
   m_parser.AddArgument(s_maxBuffersPooledPerSocket);
   m_parser.AddArgument(s_maxConsecutiveOperations);
   m_parser.AddArgument(s_recvBufferSizeArgument);
   m_parser.AddArgument(s_sendBufferSizeArgument);
   m_parser.AddArgument(s_maxPendingWritesArgument);

   m_parser.AddSwitch(s_syncConnectSwitch);
}

bool CCommandLine::Parse(
   DWORD argc,
   const char **argv)
{
   _tstring message;

   bool ok = m_parser.Parse(argc, argv, message);

   if (ok)
   {
      ok = m_parser.IsSet(s_serverArgument) && m_parser.IsSet(s_portArgument);
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxPendingWritesArgument))
      {
         if (0 != MessageRate())
         {
            message = _T("Error: -maxPendingWrites is only effective if -messageRate is set to zero");

            ok = false;
         }
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxBuffersPooledPerSocket))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxBuffersPooledPerSocket).c_str());

         if (value > std::numeric_limits<PerSocketBufferPoolingLimit>::max())
         {
            message = _T("Error: -perSocketBufferPoolSize value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxConsecutiveOperations))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxConsecutiveOperations).c_str());

         if (value > std::numeric_limits<ConsecutiveSocketOperations>::max())
         {
            message = _T("Error: -maxOps value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (!ok)
   {
      ShowHelp(message);
   }

   return ok ;
}

_tstring CCommandLine::Server() const
{
   return m_parser.GetValueAsString(s_serverArgument);
}

unsigned short CCommandLine::Port() const
{
   return static_cast<unsigned short>(StringToCount(s_portArgument, m_parser.GetValueAsString(s_portArgument).c_str()));
}

CCommandLine::Count CCommandLine::MessageSize() const
{
   return StringToCount(s_messageSizeArgument, m_parser.GetOptionalValueAsString(s_messageSizeArgument, _T("1024")).c_str());
}

Milliseconds CCommandLine::MessageRate() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_messageRateArgument, _T("1000")).c_str());
}

unsigned long CCommandLine::RecvBufferSize() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_recvBufferSizeArgument, _T("0")).c_str());
}

unsigned long CCommandLine::SendBufferSize() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_sendBufferSizeArgument, _T("0")).c_str());
}

bool CCommandLine::SyncConnect() const
{
   return m_parser.IsSet(s_syncConnectSwitch);
}

CCommandLine::Count CCommandLine::MaxPendingWrites() const
{
   return StringToCount(s_maxPendingWritesArgument, m_parser.GetOptionalValueAsString(s_maxPendingWritesArgument, _T("1")).c_str());
}

PerSocketBufferPoolingLimit CCommandLine::MaxBuffersPooledPerSocket() const
{
   return static_cast<PerSocketBufferPoolingLimit>(_ttoi(m_parser.GetOptionalValueAsString(s_maxBuffersPooledPerSocket, _T("10")).c_str()));
}

ConsecutiveSocketOperations CCommandLine::MaxConsecutiveOperationsOnOneSocket() const
{
   return static_cast<ConsecutiveSocketOperations>(_ttoi(m_parser.GetOptionalValueAsString(s_maxConsecutiveOperations, _T("10")).c_str()));
}

void CCommandLine::ShowHelp(
   const _tstring &message)
{
   cout << endl;
   cout << "DataFeed: v7.1 - Copyright (c) 2009 JetByte Limited" << endl;

   if (message.length() != 0)
   {
      cout << endl;
      cout << "Error: " << CStringConverter::TtoA(message) << endl;
      cout << endl;
      cout << "In command line: \"" << CStringConverter::TtoA(m_parser.GetParsedCommandLine()) << "\"" << endl;
      cout << endl;
   }

   cout << "Usage: DataFeed -server xxx.xxx.xxx.xxx -port xxxx" << endl;
   cout << endl;
   cout << "Command line parameters:" << endl;
   cout << " r -server                The server address to connect to (dotted ip)." << endl;
   cout << " r -port                  The port to connect to." << endl;
   cout << " o -messageSize           Size of each message." << endl;
   cout << "                          Defaults to 1024" << endl;
   cout << " o -messageRate           Delay for Y milliseconds between each message." << endl;
   cout << "                          Defaults to 1000" << endl;
   cout << " o -syncConnect           Use synchronous connect rather than ConnectEx()." << endl;
   cout << " o -recvBuffer            Allows you to set the TCP receive buffer size. This" << endl;
   cout << "                          can affect the TCP Receive Window size." << endl;
   cout << " o -sendBuffer            Allows you to set the TCP send buffer size." << endl;
   cout << " o -maxPendingWrites      Only valid if -messageRate is 0. Specifies how many" << endl;
   cout << "                          writes can be outstanding at any one time, used to" << endl;
   cout << "                          send the maximum amount of data in the shortest time" << endl;
   cout << "                          without flooding the local TCP stack and without" << endl;
   cout << "                          allowing uncontrolled non-paged pool usage." << endl;
   cout << "                          Defaults to 1." << endl;
   cout << " o -perSocketBufferPoolSize The number of I/O buffers pooled in the socket." << endl;
   cout << "                            Defaults to 10." << endl;
   cout << " o -maxOps                  The maximum number of consecutive operations" << endl;
   cout << "                            per socket. Defaults to 10." << endl;
   cout << endl;
   cout << " r = required, o = optional" << endl;
   cout << endl;
   cout << "See http://www.lenholgate.com for more details." << endl;

}

CCommandLine::Count CCommandLine::StringToCount(
   const _tstring &name,
   const _tstring &stringValue)
{
   const size_t value = _ttol(stringValue.c_str());

   const size_t maxCount = static_cast<size_t>(std::numeric_limits<CCommandLine::Count>::max());

   if (value > maxCount)
   {
      throw CException(_T("CCommandLine::StringToCount()"),
         name + _T(": Value too big: ") + ToString(value) + _T(" max is: ") + ToString(maxCount));
   }

   return static_cast<Count>(value);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
