///////////////////////////////////////////////////////////////////////////////
// File: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "StreamSocketConnectionManager.h"
#include "ConnectionCounter.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateStreamSockets.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::CWaitableCounter;
using JetByteTools::Core::ErrorCodeToErrorMessageA;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::ShutdownReceive;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::NoZeroByteRead;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// CStreamSocketConnectionManager
///////////////////////////////////////////////////////////////////////////////

CStreamSocketConnectionManager::CStreamSocketConnectionManager(
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   CConnectionCounter &connectionCounter,
   const Milliseconds transferRateMillis,
   const long maxPendingWrites,
   const unsigned long messageSize)
   :  JetByteTools::Socket::CStreamSocketConnectionManager(*this, pool, socketAllocator, bufferAllocator, NoZeroByteRead),
      m_connectionCounter(connectionCounter),
      m_transferRateMillis(transferRateMillis),
      m_pendingWrites(0),
      m_maxPendingWrites(maxPendingWrites),
      m_messageSize(messageSize),
      m_pMessageTemplate(new BYTE[messageSize]),
      m_pMessageSequence(reinterpret_cast<unsigned long *>(m_pMessageTemplate + sizeof(unsigned long))),
      m_pMessageType(reinterpret_cast<unsigned long *>(m_pMessageSequence + sizeof(unsigned long))),
      m_timerHandleIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_timerHandleIndex")))
{
   memset(m_pMessageTemplate, 0xCE, m_messageSize);

   *reinterpret_cast<unsigned long *>(m_pMessageTemplate) = m_messageSize;

   *m_pMessageSequence = 0;

   *m_pMessageType = 0;
}

CStreamSocketConnectionManager::~CStreamSocketConnectionManager()
{
   WaitForShutdownToComplete();

   delete [] m_pMessageTemplate;
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   m_connectionCounter.ConnectionEstablished();

   const IQueueTimers::Handle timerHandle = m_timerQueue.CreateTimer();

   socket.SetUserData(m_timerHandleIndex, timerHandle);

   if (socket.TryRead())
   {
      SendData(socket);
   }
}

void CStreamSocketConnectionManager::OnOutgoingConnectionFailed(
   IStreamSocket &socket,
   const IAddress & /*address*/,
   const DWORD lastError)
{
   {
      ILockableObject::Owner lock(m_lock);

      m_connectionErrors[lastError]++;
   }

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   m_connectionCounter.ConnectionFailed();
}

void CStreamSocketConnectionManager::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   IQueueTimers::Handle timerHandle = userData.GetUserData(m_timerHandleIndex);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      m_timerQueue.DestroyTimer(timerHandle);
   }

   m_connectionCounter.ConnectionComplete();
}

void CStreamSocketConnectionManager::OnReadCompleted(
   IStreamSocket & /*socket*/,
   CSmartBuffer & /*buffer*/)
{
   const _tstring message = _T("Unexpected read completed - we only send data!");

   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnWriteCompleted(
   IStreamSocket &socket,
   CSmartBuffer & /*buffer*/)
{
   if (m_transferRateMillis == 0)
   {
      const long currentPendingWrites = ::InterlockedDecrement(&m_pendingWrites);

      if (currentPendingWrites < m_maxPendingWrites)
      {
         WriteIfPossible(socket);
      }
   }
}

void CStreamSocketConnectionManager::OnWriteCompletionError(
   IStreamSocket & /*socket*/,
   CSmartBuffer & /*buffer*/,
   const DWORD lastError)
{
   ILockableObject::Owner lock(m_lock);

   m_writeCompletionErrors[lastError]++;
}

void CStreamSocketConnectionManager::OnConnectionReset(
   IStreamSocket & /*socket*/,
   const DWORD lastError)
{
   ILockableObject::Owner lock(m_lock);

   m_connectionResetErrors[lastError]++;
}

void CStreamSocketConnectionManager::OnError(
   const _tstring &message)
{
   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnMaximumConnections()
{
   const _tstring message = _T("OnMaximumConnections");

   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnTimer(
   UserData userData)
{
   IStreamSocket &socket = *reinterpret_cast<IStreamSocket *>(userData);

   SendData(socket);

   socket.Release();
}

void CStreamSocketConnectionManager::SendData(
   IStreamSocket &socket)
{
   if (m_transferRateMillis != 0)
   {
      WriteData(socket);

      const IQueueTimers::Handle timerHandle = socket.GetUserData(m_timerHandleIndex);

      socket.AddRef();

      m_timerQueue.SetTimer(timerHandle, *this, m_transferRateMillis, reinterpret_cast<UserData>(&socket));
   }
   else
   {
      WriteIfPossible(socket);
   }
}

void CStreamSocketConnectionManager::WriteIfPossible(
   IStreamSocket &socket)
{
   bool canWrite = false;

   size_t numWrites = 0;

   const size_t maxWriteBatch = m_maxPendingWrites;

   do
   {
      const long currentPendingWrites = ::InterlockedIncrement(&m_pendingWrites);

      canWrite = currentPendingWrites <= m_maxPendingWrites;

      if (canWrite)
      {
         WriteData(socket);

         numWrites++;
      }
      else
      {
         ::InterlockedDecrement(&m_pendingWrites);
      }
   }
   while (canWrite && numWrites < maxWriteBatch);
}

void CStreamSocketConnectionManager::WriteData(
   IStreamSocket &socket)
{
   (*m_pMessageSequence)++;

   *m_pMessageType = (*m_pMessageType)++ % 10;

   socket.Write(m_pMessageTemplate, m_messageSize);
}

void CStreamSocketConnectionManager::DumpErrors() const
{
   DumpErrors(_T("Connection establishment errors"), m_connectionErrors);
   DumpErrors(_T("Connections reset"), m_connectionResetErrors);
   DumpErrors(_T("Other connection errors"), m_connectionErrorErrors);
   DumpErrors(_T("Write completion errors"), m_writeCompletionErrors);

   if (m_errors.size() != 0)
   {
      cout << "Other errors" << endl;

      for (StringErrors::const_iterator it = m_errors.begin(); it != m_errors.end(); ++it)
      {
         cout << ToStringA(it->second) << " - " << CStringConverter::TtoA(it->first) << endl;
      }
   }
}

void CStreamSocketConnectionManager::DumpErrors(
   const _tstring &errorType,
   const LastErrors &lastErrors) const
{
   if (lastErrors.size() != 0)
   {
      cout << CStringConverter::TtoA(errorType) << endl;

      for (LastErrors::const_iterator it = lastErrors.begin(); it != lastErrors.end(); ++it)
      {
         cout << ToStringA(it->second) << " - " << ErrorCodeToErrorMessageA(it->first) << endl;
      }
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
