///////////////////////////////////////////////////////////////////////////////
// File: TestMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "StreamSocketConnectionManager.h"
#include "CommandLine.h"
#include "ConnectionCounter.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/ManualResetEvent.h"
#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"

#include "JetByteTools/IOTools/IOPool.h"
#include "JetByteTools/IOTools/BufferAllocator.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"

#pragma hdrstop

#include <algorithm>    // for min

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::CException;
using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CManualResetEvent;
using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;

using JetByteTools::IO::CIOPool;
using JetByteTools::IO::CBufferAllocator;

using std::cout;
using std::endl;
using std::cin;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   CGlobalErrorHandler errorHandler;

   int result = 1;

   CCommandLine commandLine;

   try
   {
      if (commandLine.Parse(argc, argv))
      {
         result = 0;

         const Milliseconds transferRate = commandLine.MessageRate();

         const CBufferAllocator::BufferSize messageSize = commandLine.MessageSize();

         const bool syncConnect = commandLine.SyncConnect();

         const CSequencedStreamSocketAllocator::SocketCount maxSocketsInPool = 10;

         const unsigned long recvBufferSize = commandLine.RecvBufferSize();

         const unsigned long sendBufferSize = commandLine.SendBufferSize();

         const CBufferAllocator::BufferSize bufferSize = commandLine.MessageSize();

         const CBufferAllocator::BufferCount maxBuffersInPool = (recvBufferSize != 0 ? std::max(static_cast<CBufferAllocator::BufferCount>(recvBufferSize / bufferSize * 1.1), static_cast<CBufferAllocator::BufferCount>(10)) :  10);

         const long maxPendingWrites = commandLine.MaxPendingWrites();

         CSequencedStreamSocketAllocator socketAllocator(
            maxSocketsInPool,
            commandLine.MaxBuffersPooledPerSocket(),
            commandLine.MaxConsecutiveOperationsOnOneSocket());

         CBufferAllocator bufferAllocator(
            bufferSize,
            maxBuffersInPool);

         CIOPool ioPool(
            0);                           // number of threads (0 = 2 x processors)

         ioPool.Start();

         CConnectionCounter connectionCounter(1);

         CStreamSocketConnectionManager connector(
            ioPool,
            socketAllocator,
            bufferAllocator,
            connectionCounter,
            transferRate,
            maxPendingWrites,
            messageSize);

         const CFullAddress address(
            commandLine.Server(),
            commandLine.Port(),
            CAddressTypeIPv4());       // default to IPv4 if there's ambiguity

         if (!connector.AsyncConnectAvailable(address.Type()) && !syncConnect)
         {
            commandLine.ShowHelp(
               _T("ConnectEx() is not available on this platform.\n")
               _T("Use -syncConnect to force synchronous connections"));

            return 1;
         }

         try
         {
            cout << "Connecting..." << endl;

            if (syncConnect)
            {
               connector.ConnectNoThrow(address, 0, recvBufferSize, sendBufferSize);
            }
            else
            {
               connector.AsyncConnectNoThrow(address, 0, recvBufferSize, sendBufferSize);
            }

            CSecurityDescriptorAllowAll sd;

            CSecurityAttributes sa(sd);

            CManualResetEvent shutdownEvent(CGlobalName(_T("JetByteToolsServerShutdown")), &sa);

            HANDLE handlesToWaitFor[2];

            handlesToWaitFor[0] = shutdownEvent.GetWaitHandle();
            handlesToWaitFor[1] = connectionCounter.GetConnectionsCompleteWaitHandle();

            DWORD waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, INFINITE);

            if (waitResult == WAIT_OBJECT_0 + 1)
            {
               cout << "Connections complete." << endl;

               if (connectionCounter.ConnectionsEstablished() == 1)
               {
                  cout << "Connected." << endl;

                  cout << "Sending data. Waiting for shutdown request." << endl;

                  handlesToWaitFor[1] = connectionCounter.GetNoConnectionsActiveWaitHandle();

                  DWORD waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, INFINITE);

                  if (waitResult != WAIT_OBJECT_0 &&
                      waitResult != WAIT_OBJECT_0 + 1)
                  {
                     throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
                  }
               }
            }
            else if (waitResult != WAIT_OBJECT_0)
            {
               throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
            }
         }
         catch(CException &e)
         {
            cout << "Exception: " << e.GetDetailsA() << endl;
         }
         catch(...)
         {
            cout << "Unexpected exception" << endl;
         }

         connector.WaitForShutdownToComplete();

         ioPool.WaitForShutdownToComplete();

         socketAllocator.ReleaseSockets();

         bufferAllocator.Flush();

         connector.DumpErrors();

         cout << "Shutdown complete" << endl;
      }
   }
   catch(const CException &e)
   {
      cout << "Exception: " << e.GetDetailsA() << endl;
   }
   catch(const char *pMsg)
   {
      cout << pMsg << endl;
   }
   catch(...)
   {
      cout << "Unexpected exception" << endl;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: TestMain.cpp
///////////////////////////////////////////////////////////////////////////////
