///////////////////////////////////////////////////////////////////////////////
// File: DistributionServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "DistributionServer.h"
#include "PerformanceCounters.h"
#include "IMonitorServerInstance.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"
#include "JetByteTools/SocketTools/StreamSocketBroadcastableConnectionCollection.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::ToString;
using JetByteTools::Core::BoolAsString;
using JetByteTools::Core::CException;
using JetByteTools::Core::GetStringLength;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;
using JetByteTools::IO::IBufferChain;

using JetByteTools::Socket::CFilteringStreamSocketServerEx;
using JetByteTools::Socket::CFlowControlStreamSocketConnectionFilter;
using JetByteTools::Socket::CFlowControlListProcessor;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::CStreamSocketBroadcastableConnectionCollection;
using JetByteTools::Socket::IMonitorStreamSocketFlowControl;
using JetByteTools::Socket::SocketBufferSize;
using JetByteTools::Socket::NoZeroByteRead;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CDistributionServer
///////////////////////////////////////////////////////////////////////////////

CDistributionServer::CDistributionServer(
   IMonitorServerInstance &monitor,
   const CFlowControlStreamSocketConnectionFilter::BufferCount maxBuffers,
   const CFlowControlStreamSocketConnectionFilter::PendingWriteCount maxPendingWrites,
   const CFlowControlListProcessor::ActionOnListFull action,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   ILimitConnections &connectionLimiter,
   CStreamSocketBroadcastableConnectionCollection &distributor)
   :  CFilteringStreamSocketServerEx(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, recvBufferSize, sendBufferSize, NoZeroByteRead, connectionLimiter),
      m_monitor(monitor),
      m_distributor(distributor),
      m_listProcessor(action),
      m_flowControlFilter(*this, socketAllocator, bufferAllocator, maxBuffers, *this, maxPendingWrites, false)
{

}

CDistributionServer::~CDistributionServer()
{
   WaitForShutdownToComplete();
}

void CDistributionServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   m_monitor.OnConnectionStarted();

   Output(_T("OnConnectionEstablished"));

   m_distributor.AddConnection(socket);

   socket.TryRead();
}

void CDistributionServer::OnConnectionClientClose(
    IStreamSocket &socket)
{
   m_distributor.RemoveConnection(socket);

   Output(_T("OnConnectionClientClose"));
}

void CDistributionServer::OnConnectionClosed(
   IStreamSocket & /*socket*/)
{
   m_monitor.OnConnectionComplete();

   Output(_T("OnConnectionClosed"));
}

#pragma JETBYTE_TODO("need to remove connections if present on close as an abort will otherwise leave the connection in the distributor...")

void CDistributionServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      // one type of message...
      // register...

      // adds this socket into the distribution list (based on message type?)

      buffer->Empty();

      socket.TryRead(buffer);
   }
   catch(const CException &e)
   {
      Output(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      Output(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CDistributionServer::OnWriteCompleted(
   IStreamSocket & /*socket*/,
   CSmartBuffer &buffer)
{
   m_monitor.BytesProcessed(buffer->GetUsed());
}

bool CDistributionServer::OnListFull(
   IStreamSocket &socket,
   const IMonitorStreamSocketFlowControl::ConnectionId id,
   IMonitorStreamSocketFlowControl &monitor,
   const IBuffer &buffer,
   IBufferChain &buffers,
   IBufferChain &discardedBuffers)
{
   // We only have the server implement this so that we can intercept the
   // list full notification and log the debug...

   Output(_T("CDistributionServer::OnListFull"));

   return m_listProcessor.OnListFull(socket, id, monitor, buffer, buffers, discardedBuffers);
}

void CDistributionServer::OnListStatus(
   IStreamSocket &socket,
   const DWORD percentageUsed)
{
   // We only have the server implement this so that we can intercept the
   // list full notification and log the debug...

   Output(_T("CDistributionServer::OnListFull"));

   m_listProcessor.OnListStatus(socket, percentageUsed);
}

void CDistributionServer::OnListEmpty(
   IStreamSocket &socket)
{
   // We only have the server implement this so that we can intercept the
   // list full notification and log the debug...

   Output(_T("CDistributionServer::OnListEmpty"));

   m_listProcessor.OnListEmpty(socket);
}

void CDistributionServer::OnBufferDelayed(
   const ConnectionId id,
   JetByteTools::IO::IBuffer &buffer)
{
   Output(_T("CDistributionServer::OnBufferDelayed: ") + ToString(id) + _T(": ") + ToString(&buffer));
}

void CDistributionServer::OnBufferSent(
   const ConnectionId id,
   JetByteTools::IO::IBuffer &buffer)
{
   Output(_T("CDistributionServer::OnBufferSent: ") + ToString(id) + _T(": ") + ToString(&buffer));
}

void CDistributionServer::OnBufferDiscarded(
   const ConnectionId id,
   JetByteTools::IO::IBuffer &buffer,
   const bool wasDelayed)
{
   Output(_T("CDistributionServer::OnBufferDiscarded: ") + BoolAsString(wasDelayed) + _T(": ") +
      ToString(id) + _T(": ") + ToString(&buffer));
}

void CDistributionServer::OnConnectionComplete(
   const ConnectionId id)
{
   Output(_T("CDistributionServer::OnConnectionComplete: ") + ToString(id));
}

void CDistributionServer::OnConnectionTerminated(
   const ConnectionId id)
{
   Output(_T("CDistributionServer::OnConnectionTerminated: ") + ToString(id));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: DistributionServer.cpp
///////////////////////////////////////////////////////////////////////////////
