///////////////////////////////////////////////////////////////////////////////
// File: FeedServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "FeedServer.h"
#include "PerformanceCounters.h"
#include "IMonitorServerInstance.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/StreamSocketBroadcastableConnectionCollection.h"

#pragma hdrstop

#include "JetByteTools/IOTools/SmartBuffer.h"

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::GetStringLength;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::CStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::CStreamSocketBroadcastableConnectionCollection;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::SocketBufferSize;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CFeedServer
///////////////////////////////////////////////////////////////////////////////

CFeedServer::CFeedServer(
   IMonitorServerInstance &monitor,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   ILimitConnections &connectionLimiter,
   CStreamSocketBroadcastableConnectionCollection &distributor)
   :  CStreamSocketServerEx(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, recvBufferSize, sendBufferSize, NoZeroByteRead, connectionLimiter),
      m_monitor(monitor),
      m_distributor(distributor)
{

}

CFeedServer::~CFeedServer()
{
   WaitForShutdownToComplete();
}

void CFeedServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   m_monitor.OnConnectionStarted();

   Output(_T("OnConnectionEstablished"));

   socket.TryRead();
}

void CFeedServer::OnConnectionClosed(
   IStreamSocket & /*socket*/)
{
   m_monitor.OnConnectionComplete();

   Output(_T("OnConnectionClosed"));
}

void CFeedServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      ProcessDataStream(socket, buffer);

      socket.TryRead(buffer);
   }
   catch(const CException &e)
   {
      Output(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      Output(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CFeedServer::ProcessDataStream(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   bool done;

   do
   {
      done = true;

      const IBuffer::BufferSize used = buffer->GetUsed();

      static const IBuffer::BufferSize minimumMessageSize = sizeof(unsigned long);

      if (used >= minimumMessageSize)
      {
         const IBuffer::BufferSize messageSize = *reinterpret_cast<const unsigned long *>(buffer->GetMemory());

         if (used == messageSize)
         {
            m_monitor.BytesProcessed(buffer->GetUsed());

            m_distributor.Broadcast(buffer.GetRef());

            buffer.Release();

            done = true;
         }
         else if (used > messageSize)
         {
            // we have a message, plus some more data

            // allocate a new buffer, copy the extra data into it and try again...

            CSmartBuffer message(buffer->SplitBufferFromFront(messageSize));

            m_monitor.BytesProcessed(message->GetUsed());

            m_distributor.Broadcast(message.GetRef());

            // loop again, we may have another complete message in there...

            done = false;
         }
         else if (messageSize > buffer->GetSize())
         {
            Output(_T("Error: Buffer too small\nExpecting: ") + ToString(messageSize) +
               _T("Got: ") + ToString(buffer->GetUsed()) + _T("\nBuffer size = ") +
               ToString(buffer->GetSize()) + _T("\nData = \n") +
               DumpData(buffer->GetMemory(), buffer->GetUsed(), 40));

            socket.Shutdown(ShutdownSend);

            // throw the rubbish away
            buffer->Empty();

            done = true;
         }
      }
   }
   while (!done);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
