///////////////////////////////////////////////////////////////////////////////
// File: PerformanceCounters.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "PerformanceCounters.h"
#include "PerfCountersApplicationName.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"

#include "JetByteTools/Win32Tools/SystemUtils.h"
#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"

#include "JetByteTools/PerfMonTools/PerformanceCounterInstaller.h"
#include "JetByteTools/PerfMonTools/SharedMemoryPerformanceMonitorCounters.h"
#include "JetByteTools/PerfMonTools/PerformanceDataSchemaFileExporter.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::CException;

using JetByteTools::Win32::GetModulePathName;
using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CSecurityAttributes;

using JetByteTools::PerfMon::CPerformanceCounterInstaller;
using JetByteTools::PerfMon::CPerformanceDataSchemaFileExporter;
using JetByteTools::PerfMon::CSharedMemoryPerformanceMonitorCounters;
using JetByteTools::PerfMon::CPerformanceMonitorCountersBase;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static CPerformanceMonitorCountersBase *CreateCountersIfRequired(
   bool createCounters,
   const _tstring &applicationName,
   const CPerformanceDataSchema &schema,
   const CSecurityAttributes &sa);

///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////

static const CPerformanceMonitorCountersBase::ObjectId s_noObject;

static const CPerformanceMonitorCountersBase::CounterId s_noCounter;

///////////////////////////////////////////////////////////////////////////////
// CPerformanceCounters::CounterCollection
///////////////////////////////////////////////////////////////////////////////

CPerformanceCounters::CounterCollection::CounterCollection(
   CPerformanceMonitorCountersBase *pCounters,
   const _tstring &instanceName,
   const bool instancePersists)
   :  m_object(pCounters ? pCounters->AllocatePerformanceObjectInstance(CPerformanceDataSchema::s_serverExeObjectName, instanceName) : s_noObject),
      m_bytesProcessedCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_bytesProcessedCounter) : s_noCounter),
      m_bytesPerSecondCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_bytesPerSecondCounter) : s_noCounter),
      m_totalBuffersCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_totalBuffersCounter) : s_noCounter),
      m_buffersInUseCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_buffersInUseCounter) : s_noCounter),
      m_connectionsActiveCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_connectionsActiveCounter) : s_noCounter),
      m_totalSocketsCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_totalSocketsCounter) : s_noCounter),
      m_socketsInUseCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_socketsInUseCounter) : s_noCounter),
      m_ioThreadsActiveCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_ioThreadsActiveCounter) : s_noCounter),
      m_ioThreadsProcessingCounter(pCounters ? pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_ioThreadsProcessingCounter) : s_noCounter),
      m_pCounters(pCounters),
      m_instancePersists(instancePersists)
{
}

CPerformanceCounters::CounterCollection::~CounterCollection()
{
   ReleaseCounters();
}

void CPerformanceCounters::CounterCollection::ReleaseCounters()
{
   if (!m_instancePersists && m_pCounters)
   {
      m_pCounters->ReleasePerformanceObjectInstance(m_object);
   }

   m_pCounters = 0;
}

///////////////////////////////////////////////////////////////////////////////
// CPerformanceCounters::ServerInstanceMonitor
///////////////////////////////////////////////////////////////////////////////

class CPerformanceCounters::ServerInstanceMonitor : public IMonitorServerInstance
{
   public :

      ServerInstanceMonitor(
         const _tstring &instanceName,
         IMonitorServerInstance &parent,
         const CPerformanceMonitorCountersBase::ObjectId &parentObject,
         CPerformanceMonitorCountersBase *pCounters);

      ServerInstanceMonitor(
         const ServerInstanceMonitor &rhs) = delete;

      ~ServerInstanceMonitor() override;

      ServerInstanceMonitor &operator=(
         const ServerInstanceMonitor &rhs) = delete;

   private :

      // Implement IMonitorServerInstance

      void BytesProcessed(
         const long bytes) override;

      void OnConnectionStarted() override;
      void OnConnectionComplete() override;

      IMonitorServerInstance &m_parent;

      CPerformanceMonitorCountersBase *m_pCounters;

      const CPerformanceMonitorCountersBase::ObjectId m_object;

      const CPerformanceMonitorCountersBase::CounterId m_bytesProcessedCounter;
      const CPerformanceMonitorCountersBase::CounterId m_bytesPerSecondCounter;
      const CPerformanceMonitorCountersBase::CounterId m_connectionsActiveCounter;
};

///////////////////////////////////////////////////////////////////////////////
// CPerformanceCounters
///////////////////////////////////////////////////////////////////////////////

CPerformanceCounters::CPerformanceCounters(
   const bool createCounters,
   const _tstring &instanceName,
   CSecurityAttributes &sa)
   :  m_pCounters(CreateCountersIfRequired(createCounters, s_applicationName, m_schema, sa)),
      m_serverInstance(m_pCounters, instanceName, false),
      m_totalInstance(m_pCounters, _T("_Total"), true)
{
}

CPerformanceCounters::~CPerformanceCounters()
{
   m_serverInstance.ReleaseCounters();

   m_totalInstance.ReleaseCounters();

   for (ServerInstanceMonitors::iterator it = m_serverInstanceMonitors.begin(); it != m_serverInstanceMonitors.end(); ++it)
   {
      delete it->second;
   }

   delete m_pCounters;
}

bool CPerformanceCounters::InstallCounters()
{
   bool installed = false;

   CPerformanceCounterInstaller installer;

   if (!installer.CountersAreInstalled(s_applicationName))
   {
      const _tstring counterPath = GetModulePathName() + _T("\\DistributionServerCounters.dll");

      CPerformanceDataSchema schema;

      CPerformanceDataSchemaFileExporter exporter;

      installer.Install(
         s_applicationName,
         counterPath,
         false,
         exporter.ExportHeaderFile(s_applicationName, schema),
         exporter.ExportIniFile(s_applicationName, schema),
         _T(""),
         10000,
         10000,
         schema.GetChecksum());

      installed = true;
   }

   return installed;
}

bool CPerformanceCounters::RemoveCounters()
{
   bool removed = false;

   CPerformanceCounterInstaller installer;

   if (installer.CountersAreInstalled(s_applicationName))
   {
      installer.Uninstall(s_applicationName);

      removed = true;
   }

   return removed;
}

IMonitorServerInstance &CPerformanceCounters::GetServerInstanceMonitor(
   const _tstring &instanceName)
{
   ServerInstanceMonitor *pMonitor = 0;

   ServerInstanceMonitors::const_iterator it = m_serverInstanceMonitors.find(instanceName);

   if (it != m_serverInstanceMonitors.end())
   {
      pMonitor = it->second;
   }

   if (!pMonitor)
   {
      pMonitor = new ServerInstanceMonitor(instanceName, *this, m_serverInstance.m_object, m_pCounters);

      m_serverInstanceMonitors[instanceName] = pMonitor;
   }

   return *pMonitor;
}

void CPerformanceCounters::OnSocketCreated()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_totalSocketsCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_totalSocketsCounter);
   }
}

void CPerformanceCounters::OnSocketAttached(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_socketsInUseCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_socketsInUseCounter);
   }
}

void CPerformanceCounters::OnConnectionStarted()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_connectionsActiveCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_connectionsActiveCounter);
   }
}

void CPerformanceCounters::OnConnectionComplete()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_serverInstance.m_connectionsActiveCounter);
      m_pCounters->DecrementPerformanceCounter32(m_totalInstance.m_connectionsActiveCounter);
   }
}

void CPerformanceCounters::OnSocketReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_serverInstance.m_socketsInUseCounter);
      m_pCounters->DecrementPerformanceCounter32(m_totalInstance.m_socketsInUseCounter);
   }
}

void CPerformanceCounters::OnSocketDestroyed()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_serverInstance.m_totalSocketsCounter);
      m_pCounters->DecrementPerformanceCounter32(m_totalInstance.m_totalSocketsCounter);
   }
}

void CPerformanceCounters::BytesProcessed(
   const long bytes)
{
   if (m_pCounters)
   {
      m_pCounters->AddToPerformanceCounter64(m_serverInstance.m_bytesProcessedCounter, bytes);
      m_pCounters->AddToPerformanceCounter64(m_totalInstance.m_bytesProcessedCounter, bytes);

      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_bytesPerSecondCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_bytesPerSecondCounter);
   }
}

void CPerformanceCounters::OnBufferCreated()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_totalBuffersCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_totalBuffersCounter);
   }
}

void CPerformanceCounters::OnBufferDestroyed()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_serverInstance.m_totalBuffersCounter);
      m_pCounters->DecrementPerformanceCounter32(m_totalInstance.m_totalBuffersCounter);
   }
}

void CPerformanceCounters::OnBufferAllocated(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_buffersInUseCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_buffersInUseCounter);
   }
}

void CPerformanceCounters::OnBufferAllocationContention()
{
}

void CPerformanceCounters::OnBufferReleaseContention()
{
}

void CPerformanceCounters::OnBufferReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_serverInstance.m_buffersInUseCounter);
      m_pCounters->DecrementPerformanceCounter32(m_totalInstance.m_buffersInUseCounter);
   }
}

void CPerformanceCounters::OnIOPoolError(
   const _tstring & /*message*/)
{
}

void CPerformanceCounters::OnIOPoolThreadCreated()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_ioThreadsActiveCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_ioThreadsActiveCounter);
   }
}

void CPerformanceCounters::OnIOPoolDispatch()
{
}

void CPerformanceCounters::OnIOPoolThreadDestroyed()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_serverInstance.m_ioThreadsActiveCounter);
      m_pCounters->DecrementPerformanceCounter32(m_totalInstance.m_ioThreadsActiveCounter);
   }
}

void CPerformanceCounters::OnIOPoolThreadBeginProcessing()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_serverInstance.m_ioThreadsProcessingCounter);
      m_pCounters->IncrementPerformanceCounter32(m_totalInstance.m_ioThreadsProcessingCounter);
  }
}

void CPerformanceCounters::OnIOPoolThreadEndProcessing()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_serverInstance.m_ioThreadsProcessingCounter);
      m_pCounters->DecrementPerformanceCounter32(m_totalInstance.m_ioThreadsProcessingCounter);
   }
}

///////////////////////////////////////////////////////////////////////////////
// CPerformanceCounters::ServerInstanceMonitor
///////////////////////////////////////////////////////////////////////////////

CPerformanceCounters::ServerInstanceMonitor::ServerInstanceMonitor(
   const _tstring &instanceName,
   IMonitorServerInstance &parent,
   const CPerformanceMonitorCountersBase::ObjectId &parentObject,
   CPerformanceMonitorCountersBase *pCounters)
   :  m_parent(parent),
      m_pCounters(pCounters),
      m_object(m_pCounters ? m_pCounters->AllocateChildPerformanceObjectInstance(
         CPerformanceDataSchema::s_serverInstanceObjectName,
         instanceName,
         parentObject) : s_noObject),
      m_bytesProcessedCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_bytesProcessedCounter) : s_noCounter),
      m_bytesPerSecondCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_bytesPerSecondCounter) : s_noCounter),
      m_connectionsActiveCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, CPerformanceDataSchema::s_connectionsActiveCounter) : s_noCounter)
{

}

CPerformanceCounters::ServerInstanceMonitor::~ServerInstanceMonitor()
{
   if (m_pCounters)
   {
      m_pCounters->ReleasePerformanceObjectInstance(m_object);
   }
}

void CPerformanceCounters::ServerInstanceMonitor::BytesProcessed(
   const long bytes)
{
   if (m_pCounters)
   {
      m_pCounters->AddToPerformanceCounter64(m_bytesProcessedCounter, bytes);

      m_pCounters->IncrementPerformanceCounter32(m_bytesPerSecondCounter);

      m_parent.BytesProcessed(bytes);
   }
}

void CPerformanceCounters::ServerInstanceMonitor::OnConnectionStarted()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_connectionsActiveCounter);

      m_parent.OnConnectionStarted();
   }

}

void CPerformanceCounters::ServerInstanceMonitor::OnConnectionComplete()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_connectionsActiveCounter);

      m_parent.OnConnectionComplete();
   }
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static CPerformanceMonitorCountersBase *CreateCountersIfRequired(
   const bool createCounters,
   const _tstring &applicationName,
   const CPerformanceDataSchema &schema,
   const CSecurityAttributes &sa)
{
   CPerformanceMonitorCountersBase *pCounters = nullptr;

   if (createCounters)
   {
      if (CPerformanceCounterInstaller().CountersAreInstalled(applicationName))
      {
         if (CPerformanceCounterInstaller().GetSchemaChecksum(applicationName) != schema.GetChecksum())
         {
            throw CException(
               _T("CPerformanceCounters::CreateCountersIfRequired()"),
               _T("Performance counter schema mismatch; these counters aren't the ones that are installed."));
         }

         const CGlobalName name(applicationName);

         const DWORD firstCounterIndex = CPerformanceCounterInstaller().FirstCounterIndex(s_applicationName);

         pCounters = new CSharedMemoryPerformanceMonitorCounters(name, schema, firstCounterIndex, const_cast<CSecurityAttributes *>(&sa));
      }
      else
      {
         //OutputEx(_T("Not using performance counters as they are not currently installed. Run the service with /InstallCounters to install them."));
      }
   }

   return pCounters;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: PerformanceCounters.cpp
///////////////////////////////////////////////////////////////////////////////
