///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "FeedServer.h"
#include "DistributionServer.h"
#include "PerformanceCounters.h"
#include "CommandLine.h"

#include "ServerCommon/SimpleServerShutdownHandler.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/Win32Tools/ManualResetEvent.h"
#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"

#include "JetByteTools/IOTools/IOPool.h"
#include "JetByteTools/IOTools/BufferAllocator.h"
#include "JetByteTools/IOTools/BufferHandleAllocator.h"
#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"
#include "JetByteTools/SocketTools/ServerCollection.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/StreamSocketBroadcastableConnectionCollection.h"
#include "JetByteTools/SocketTools/FlowControlListProcessor.h"
#include "JetByteTools/SocketTools/FlowControlStreamSocketConnectionFilter.h"
#include "JetByteTools/SocketTools/FullAddress.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;

using JetByteTools::IO::CIOPool;
using JetByteTools::IO::CBufferAllocator;
using JetByteTools::IO::CBufferHandleAllocator;
using JetByteTools::IO::CAsyncFileLog;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::CServerCollection;
using JetByteTools::Socket::IIPAddress;
using JetByteTools::Socket::CStreamSocketBroadcastableConnectionCollection;
using JetByteTools::Socket::CFlowControlListProcessor;
using JetByteTools::Socket::CFlowControlStreamSocketConnectionFilter;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("DistributionServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CCommandLineEx commandLine(argc, argv, _T("DistributionServer"));

         if (commandLine.Parse())
         {
            if (commandLine.RegisterCounters())
            {
               if (CPerformanceCounters::InstallCounters())
               {
                  OutputEx(_T("Counters installed"));
               }
               else
               {
                  OutputEx(_T("Counters already installed"));
               }
            }
            else if (commandLine.UnRegisterCounters())
            {
               if (CPerformanceCounters::RemoveCounters())
               {
                  OutputEx(_T("Counters Removed"));
               }
               else
               {
                  OutputEx(_T("Counters not installed"));
               }
            }
            else
            {
               CSecurityDescriptorAllowAll sd;

               CSecurityAttributes sa(sd);

               const _tstring pid = ToString(GetCurrentProcessId());

               const _tstring instanceName = _T("Server:") + pid;

               CPerformanceCounters counters(true, instanceName, sa);

               CIOPool pool(
                  counters,
                  commandLine.NumberOfIOThreads());   // number of threads (0 = 2 x processors)

               pool.Start();

               CSequencedStreamSocketAllocator socketAllocator(
                  counters,
                  commandLine.SocketPoolSize(),
                  commandLine.MaxBuffersPooledPerSocket(),
                  commandLine.MaxConsecutiveOperationsOnOneSocket());

               CBufferAllocator bufferAllocator(
                  counters,
                  commandLine.BufferSize(),
                  commandLine.BufferPoolSize());

               CBufferHandleAllocator handleAllocator(
                  counters,
                  commandLine.BufferPoolSize(),
                  bufferAllocator);

               CStreamSocketBroadcastableConnectionCollection distributor(handleAllocator);

               const ListenBacklog listenBacklog = commandLine.ListenBacklog();

               CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

               CServerCollection servers;

               CFullAddress feedAddress(
                  commandLine.Server(),
                  commandLine.Port(),
                  CAddressTypeIPv4());       // default to IPv4 if there's ambiguity

               servers.AddServer(
                  new CFeedServer(
                  counters.GetServerInstanceMonitor(_T("Feed:") + ToString(commandLine.Port())),
                  feedAddress,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  commandLine.RecvBufferSize(),
                  commandLine.SendBufferSize(),
                  connectionLimiter,
                  distributor));

               CFullAddress distributionAddress(
                  commandLine.Server(),
                  static_cast<IIPAddress::Port>(commandLine.Port() + 1),
                  CAddressTypeIPv4());

               const CFlowControlStreamSocketConnectionFilter::BufferCount maxBuffers = commandLine.MaxBuffersQueued();

               const CFlowControlStreamSocketConnectionFilter::PendingWriteCount maxPendingWrites = commandLine.MaxPendingWrites();

               CFlowControlListProcessor::ActionOnListFull action = commandLine.ActionOnBufferQueueFull();

               servers.AddServer(
                  new CDistributionServer(
                  counters.GetServerInstanceMonitor(_T("Dist:") + ToString(commandLine.Port() + 1)),
                  maxBuffers,
                  maxPendingWrites,
                  action,
                  distributionAddress,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  commandLine.RecvBufferSize(),
                  commandLine.SendBufferSize(),
                  connectionLimiter,
                  distributor));

               servers.Start();

               servers.StartAcceptingConnections();

               shutdownHandler.SetServerControl(servers, commandLine.EventNamePrefix());

               shutdownHandler.WaitForShutdownRequest();

               servers.WaitForShutdownToComplete();

               pool.WaitForShutdownToComplete();

               socketAllocator.ReleaseSockets();

               bufferAllocator.Flush();
            }
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
