///////////////////////////////////////////////////////////////////////////////
// File: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "StreamSocketConnectionManager.h"
#include "ConnectionCounter.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/WaitableCounter.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateStreamSockets.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ErrorCodeToErrorMessageA;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::ToString;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::NoZeroByteRead;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// CStreamSocketConnectionManager
///////////////////////////////////////////////////////////////////////////////

CStreamSocketConnectionManager::CStreamSocketConnectionManager(
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter,
   CConnectionCounter &connectionCounter)
   :  JetByteTools::Socket::CStreamSocketConnectionManager(*this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_connectionCounter(connectionCounter)
{

}

CStreamSocketConnectionManager::~CStreamSocketConnectionManager()
{
   WaitForShutdownToComplete();
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   m_connectionCounter.ConnectionEstablished();

   socket.TryRead();
}

void CStreamSocketConnectionManager::OnOutgoingConnectionFailed(
   IStreamSocket & /*socket*/,
   const IAddress & /*address*/,
   const DWORD lastError)
{
   m_connectionErrors[lastError]++;

   m_connectionCounter.ConnectionFailed();
}

void CStreamSocketConnectionManager::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      ProcessData(socket, buffer);

      socket.TryRead(buffer);
   }
   catch(const CException &e)
   {
      OutputEx(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CStreamSocketConnectionManager::OnConnectionReset(
   IStreamSocket & /*socket*/,
   const DWORD lastError)
{
   m_connectionResetErrors[lastError]++;
}

void CStreamSocketConnectionManager::OnError(
   const _tstring &message)
{
   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnMaximumConnections()
{
   const _tstring message = _T("OnMaximumConnections");

   m_errors[message]++;
}

void CStreamSocketConnectionManager::ProcessData(
   IStreamSocket &/*socket*/,
   CSmartBuffer &buffer)
{
   // do the usual length prefixed stuff

   buffer->Empty();
}

void CStreamSocketConnectionManager::DumpErrors() const
{
   DumpErrors(_T("Connection establishment errors"), m_connectionErrors);
   DumpErrors(_T("Connections reset"), m_connectionResetErrors);
   DumpErrors(_T("Other connection errors"), m_connectionErrorErrors);

   if (m_errors.size() != 0)
   {
      cout << "Other errors" << endl;

      for (StringErrors::const_iterator it = m_errors.begin(); it != m_errors.end(); ++it)
      {
         cout << ToStringA(it->second) << " - " << CStringConverter::TtoA(it->first) << endl;
      }
   }
}

void CStreamSocketConnectionManager::DumpErrors(
   const _tstring &errorType,
   const LastErrors &lastErrors) const
{
   if (lastErrors.size() != 0)
   {
      cout << CStringConverter::TtoA(errorType) << endl;

      for (LastErrors::const_iterator it = lastErrors.begin(); it != lastErrors.end(); ++it)
      {
         cout << ToStringA(it->second) << " - " << ErrorCodeToErrorMessageA(it->first) << endl;
      }
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
