///////////////////////////////////////////////////////////////////////////////
// File: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2005 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ToString.h"
#include "JetByteTools/CoreTools/Tchar.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToString;

using JetByteTools::IO::IBuffer;

using JetByteTools::Socket::CAddressIPv4;
using JetByteTools::Socket::ZeroByteReadConfiguration;
using JetByteTools::Socket::ZeroByteSyncRead;
using JetByteTools::Socket::ZeroByteAsyncRead;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::PerSocketBufferPoolingLimit;
using JetByteTools::Socket::ConsecutiveSocketOperations;

using JetByteTools::Socket::CFlowControlStreamSocketConnectionFilter;

using std::string;
using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Constants
///////////////////////////////////////////////////////////////////////////////

static const _tstring s_serverArgument                   = _T("server");
static const _tstring s_portArgument                     = _T("port");
static const _tstring s_numberOfIOThreadsArgument        = _T("numberOfIOThreads");
static const _tstring s_listenBacklogArgument            = _T("listenBacklog");
static const _tstring s_socketPoolSizeArgument           = _T("socketPoolSize");
static const _tstring s_bufferPoolSizeArgument           = _T("bufferPoolSize");
static const _tstring s_bufferPoolDisabledSwitch         = _T("noBufferPool");
static const _tstring s_bufferSizeArgument               = _T("bufferSize");
static const _tstring s_maxConnectionsArgument           = _T("maxConnections");
static const _tstring s_displayDebugSwitch               = _T("displayDebug");
static const _tstring s_noWelcomeMessageSwitch           = _T("noWelcomeMessage");
static const _tstring s_registerCountersSwitch           = _T("registerCounters");
static const _tstring s_unRegisterCountersSwitch         = _T("unRegisterCounters");
static const _tstring s_multicastAddressArgument         = _T("multicastAddress");
static const _tstring s_recvBufferSizeArgument           = _T("recvBuffer");
static const _tstring s_sendBufferSizeArgument           = _T("sendBuffer");
static const _tstring s_syncZeroByteReadSwitch           = _T("syncZero");
static const _tstring s_asyncZeroByteReadSwitch          = _T("asyncZero");
//static const _tstring s_pageAlignedBuffersSwitch         = _T("pageAlignedBuffers");
//static const _tstring s_pageSizedBuffersSwitch           = _T("pageSizedBuffers");
static const _tstring s_numberOfEntriesPerCallArgument   = _T("numEntries");
static const _tstring s_maxBuffersPooledPerSocket        = _T("perSocketBufferPoolSize");
static const _tstring s_maxConsecutiveOperations         = _T("maxOps");

static const _tstring s_connectionTimeoutArgument        = _T("connectionTimeout");
static const _tstring s_eventNamePrefixArgument          = _T("eventNamePrefix");
static const _tstring s_maxBuffersArgument               = _T("maxBuffers");
static const _tstring s_maxPendingWritesArgument         = _T("maxPendingWrites");
static const _tstring s_notificationPercentageArgument   = _T("notify");
static const _tstring s_processingDelayArgument          = _T("delay");

///////////////////////////////////////////////////////////////////////////////
// CCommandLine
///////////////////////////////////////////////////////////////////////////////

CCommandLine::CCommandLine(
   const DWORD argc,
   const char **argv,
   const _tstring &serverName,
   const DWORD type)
   :  m_argc(argc),
      m_argv(argv),
      m_serverName(serverName),
      m_type(static_cast<ServerType>(type))
{
   m_parser.AddArgument(s_serverArgument);
   m_parser.AddArgument(s_portArgument);

   m_parser.AddArgument(s_numberOfIOThreadsArgument);

   if (m_type & TCPServer || m_type & UDPServer)
   {
      m_parser.AddArgument(s_listenBacklogArgument);
      m_parser.AddArgument(s_maxConnectionsArgument);
   }

   m_parser.AddArgument(s_socketPoolSizeArgument);
   m_parser.AddArgument(s_bufferPoolSizeArgument);
   m_parser.AddArgument(s_bufferSizeArgument);

   m_parser.AddSwitch(s_bufferPoolDisabledSwitch);

   if (m_type & SupportsSendRecvBufSize)
   {
      m_parser.AddArgument(s_recvBufferSizeArgument);
      m_parser.AddArgument(s_sendBufferSizeArgument);
   }

   if (m_type & TCPServer)
   {
      if ((m_type & DoesNotSupportWelcomeMessage) == 0)
      {
         m_parser.AddSwitch(s_noWelcomeMessageSwitch);
      }
   }

   if (m_type & SupportsPerfCounters)
   {
      m_parser.AddSwitch(s_registerCountersSwitch);
      m_parser.AddSwitch(s_unRegisterCountersSwitch);
   }

   if ((m_type & UDPMulticastServer) == UDPMulticastServer)
   {
      m_parser.AddArgument(s_multicastAddressArgument);
   }

   if (m_type & SupportsZeroByteRead)
   {
      m_parser.AddSwitch(s_syncZeroByteReadSwitch);
      m_parser.AddSwitch(s_asyncZeroByteReadSwitch);
   }

   //if (m_type & SupportsPageAllocator)
   //{
   //   m_parser.AddSwitch(s_pageAlignedBuffersSwitch);
   //   m_parser.AddSwitch(s_pageSizedBuffersSwitch);
   //}

   if ((m_type & UDPConnectionServer) == UDPConnectionServer)
   {
      m_parser.AddArgument(s_connectionTimeoutArgument);
   }

   DEBUG_ONLY(m_parser.AddSwitch(s_displayDebugSwitch));

   m_parser.AddArgument(s_eventNamePrefixArgument);

   if ((m_type & SupportsFlowControl) == SupportsFlowControl)
   {
      m_parser.AddArgument(s_maxBuffersArgument);
      m_parser.AddArgument(s_maxPendingWritesArgument);
      m_parser.AddArgument(s_notificationPercentageArgument);
   }

   if ((m_type & SupportsProcessingDelay) == SupportsProcessingDelay)
   {
      m_parser.AddArgument(s_processingDelayArgument);
   }

   if ((m_type & SupportsIOPoolEx) == SupportsIOPoolEx)
   {
      m_parser.AddArgument(s_numberOfEntriesPerCallArgument);
   }

   m_parser.AddArgument(s_maxBuffersPooledPerSocket);
   m_parser.AddArgument(s_maxConsecutiveOperations);
}

bool CCommandLine::Parse()
{
   _tstring message;

   bool ok = m_parser.Parse(m_argc, m_argv, message);

   if (ok)
   {
      if (m_type & SupportsPerfCounters)
      {
         if (RegisterCounters() && UnRegisterCounters())
         {
            ok = false;

            message = _T("only one of -registerCounters and -unRegisterCounters can be specified!");
         }
         else if (RegisterCounters() || UnRegisterCounters())
         {
            return ok;
         }
      }
   }

   if (ok)
   {
      ok = m_parser.IsSet(s_portArgument);

      message = _T("-port is required!");
   }

   if (ok)
   {
      if (m_parser.IsSet(s_bufferPoolDisabledSwitch) && m_parser.IsSet(s_bufferPoolSizeArgument))
      {
         ok = false;

         message = _T("it does not make sense to set -bufferPoolSize when -noBufferPool is set!");
      }
   }

   if (ok)
   {
      if (m_type & SupportsZeroByteRead)
      {
         if (m_parser.IsSet(s_syncZeroByteReadSwitch) &&
             m_parser.IsSet(s_asyncZeroByteReadSwitch))
         {
            ok = false;

            message = _T("only one of -syncZero and -asyncZero can be specified!");
         }
      }
   }

   if (ok)
   {
      //if (m_type & SupportsPageAllocator)
      //{
      //   if (m_parser.IsSet(s_pageSizedBuffersSwitch))
      //   {
      //      const IBuffer::BufferSize bufferSize = BufferSize();

      //      // Arbitrary limit, this is 128k on a 4k page machine...

      //      if (bufferSize > 32)
      //      {
      //         message = _T("buffer size is in pages when -pageSizedBuffers is specified, -bufferSize ") +
      //            ToString(bufferSize) + _T(" pages per buffer is too much! Max is 32!");

      //         ok = false;
      //      }
      //   }
      //}
   }

   if (ok)
   {
      if ((m_type & SupportsFlowControl) == SupportsFlowControl)
      {
         if (m_parser.IsSet(s_maxBuffersArgument))
         {
            const size_t value = _ttoi(m_parser.GetOptionalValueAsString(s_maxBuffersArgument, _T("0")).c_str());

            const size_t maxValue = std::numeric_limits<DWORD>::max();

            if (value > maxValue)
            {
               message = ToString(value) + _T(" is too big for -maxBuffers. Max is ") + ToString(maxValue) + _T("!");

               ok = false;
            }
         }
      }
   }

   if (ok)
   {
      if ((m_type & SupportsFlowControl) == SupportsFlowControl)
      {
         if (m_parser.IsSet(s_notificationPercentageArgument))
         {
            const int value = _ttoi(m_parser.GetOptionalValueAsString(s_notificationPercentageArgument, _T("0")).c_str());

            if (value > 100)
            {
               message = ToString(value) + _T(" is too big for -notify. Max is 100!");

               ok = false;
            }
            else if (value == 0)
            {
               message = _T("-notify must be followed by a number between 1 and 100!");

               ok = false;
            }
         }
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxBuffersPooledPerSocket))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxBuffersPooledPerSocket).c_str());

         if (value > std::numeric_limits<PerSocketBufferPoolingLimit>::max())
         {
            message = _T("Error: -perSocketBufferPoolSize value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxConsecutiveOperations))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxConsecutiveOperations).c_str());

         if (value > std::numeric_limits<ConsecutiveSocketOperations>::max())
         {
            message = _T("Error: -maxOps value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (!ok)
   {
      ShowHelp(message);
   }

   return ok ;
}

_tstring CCommandLine::EventNamePrefix() const
{
   return m_parser.GetOptionalValueAsString(s_eventNamePrefixArgument, _T("JetByteTools"));
}

_tstring CCommandLine::Server() const
{
   if (m_type & TCPClient)
   {
      return m_parser.GetOptionalValueAsString(s_serverArgument, _T("localhost"));
   }
   else
   {
      return m_parser.GetOptionalValueAsString(s_serverArgument, _T("0.0.0.0"));
   }
}

CAddressIPv4::Port CCommandLine::Port() const
{
   return static_cast<unsigned short>(_ttol(m_parser.GetValueAsString(s_portArgument).c_str()));
}

CCommandLine::Count CCommandLine::NumberOfIOThreads() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_numberOfIOThreadsArgument, _T("0")).c_str());
}

bool CCommandLine::DisplayDebug() const
{
   if (m_parser.IsSwitch(s_displayDebugSwitch))
   {
      return m_parser.IsSet(s_displayDebugSwitch);
   }

   return false;
}

JetByteTools::Socket::ListenBacklog CCommandLine::ListenBacklog() const
{
   return static_cast<JetByteTools::Socket::ListenBacklog>(_ttol(m_parser.GetOptionalValueAsString(s_listenBacklogArgument, _T("5")).c_str()));
}

CCommandLine::Count CCommandLine::SocketPoolSize() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_socketPoolSizeArgument, _T("10")).c_str());
}

JetByteTools::IO::CBufferAllocator::BufferCount CCommandLine::BufferPoolSize() const
{
   if (m_parser.IsSet(s_bufferPoolDisabledSwitch))
   {
      return JetByteTools::IO::CBufferAllocator::NoPooledBuffers;
   }

   return StringToUnsignedLong(m_parser.GetOptionalValueAsString(s_bufferPoolSizeArgument, _T("10")).c_str());
}

IBuffer::BufferSize CCommandLine::BufferSize() const
{
   //if (m_type & SupportsPageAllocator)
   //{
   //   if (m_parser.IsSet(s_pageSizedBuffersSwitch))
   //   {
   //      return StringToUnsignedLong(m_parser.GetOptionalValueAsString(s_bufferSizeArgument, _T("1")).c_str());
   //   }
   //}

   return StringToUnsignedLong(m_parser.GetOptionalValueAsString(s_bufferSizeArgument, _T("1024")).c_str());
}

CCommandLine::Count CCommandLine::MaxConnections() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_maxConnectionsArgument, _T("0")).c_str());
}

bool CCommandLine::NoWelcomeMessage() const
{
   return m_parser.IsSet(s_noWelcomeMessageSwitch);
}

bool CCommandLine::RegisterCounters() const
{
   return m_parser.IsSet(s_registerCountersSwitch);
}

bool CCommandLine::UnRegisterCounters() const
{
   return m_parser.IsSet(s_unRegisterCountersSwitch);
}

bool CCommandLine::HasMulticastAddress() const
{
   return m_parser.IsSet(s_multicastAddressArgument);
}

_tstring CCommandLine::MulticastAddress() const
{
   return m_parser.GetOptionalValueAsString(s_multicastAddressArgument, _T("0.0.0.0"));
}

unsigned long CCommandLine::RecvBufferSize() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_recvBufferSizeArgument, _T("0")).c_str());
}

unsigned long CCommandLine::SendBufferSize() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_sendBufferSizeArgument, _T("0")).c_str());
}

ZeroByteReadConfiguration CCommandLine::GetZeroByteReadConfiguration() const
{
   if (m_parser.IsSet(s_syncZeroByteReadSwitch))
   {
      return ZeroByteSyncRead;
   }

   if (m_parser.IsSet(s_asyncZeroByteReadSwitch))
   {
      return ZeroByteAsyncRead;
   }

   return NoZeroByteRead;
}

JetByteTools::IO::CBufferAllocator::AllocationFlags CCommandLine::GetAllocationFlags() const
{
   int flags = JetByteTools::IO::CBufferAllocator::SizeInBytes;

   //if (m_type & SupportsPageAllocator)
   //{
   //   if (m_parser.IsSet(s_pageSizedBuffersSwitch))
   //   {
   //      flags = JetByteTools::IO::CBufferAllocator::SizeInPages;
   //   }

   //   if (m_parser.IsSet(s_pageAlignedBuffersSwitch))
   //   {
   //      flags |= JetByteTools::IO::CBufferAllocator::UsePageAllocator;
   //   }
   //}

   return static_cast<JetByteTools::IO::CBufferAllocator::AllocationFlags>(flags);
}

Milliseconds CCommandLine::ConnectionTimeout() const
{
   return _ttoi(m_parser.GetOptionalValueAsString(s_connectionTimeoutArgument, _T("10000")).c_str());
}

CFlowControlStreamSocketConnectionFilter::BufferCount CCommandLine::MaxBuffers() const
{
   return static_cast<CFlowControlStreamSocketConnectionFilter::BufferCount>(_ttoi(m_parser.GetOptionalValueAsString(s_maxBuffersArgument, _T("10")).c_str()));
}

CFlowControlStreamSocketConnectionFilter::PendingWriteCount CCommandLine::MaxPendingWrites() const
{
   return _ttoi(m_parser.GetOptionalValueAsString(s_notificationPercentageArgument, _T("10")).c_str());
}

CFlowControlStreamSocketConnectionFilter::Percentage CCommandLine::NotificationPercentageStepSize() const
{
   return static_cast<BYTE>(_ttoi(m_parser.GetOptionalValueAsString(s_notificationPercentageArgument, _T("25")).c_str()));
}

Milliseconds CCommandLine::ProcessingDelay() const
{
   return _ttoi(m_parser.GetOptionalValueAsString(s_processingDelayArgument, _T("0")).c_str());
}

DWORD CCommandLine::NumberOfEntriesPerCall() const
{
   return _ttoi(m_parser.GetOptionalValueAsString(s_numberOfEntriesPerCallArgument, _T("1")).c_str());
}

PerSocketBufferPoolingLimit CCommandLine::MaxBuffersPooledPerSocket() const
{
   return static_cast<PerSocketBufferPoolingLimit>(_ttoi(m_parser.GetOptionalValueAsString(s_maxBuffersPooledPerSocket, _T("10")).c_str()));
}

ConsecutiveSocketOperations CCommandLine::MaxConsecutiveOperationsOnOneSocket() const
{
   return static_cast<ConsecutiveSocketOperations>(_ttoi(m_parser.GetOptionalValueAsString(s_maxConsecutiveOperations, _T("10")).c_str()));
}

void CCommandLine::ShowHelp(
   const _tstring &message)
{
   const string serverName = CStringConverter::TtoA(m_serverName);

   cout << endl;
   cout << serverName << " - v7.0 - Copyright (c) 2015 JetByte Limited" << endl;

   if (message.length() != 0)
   {
      cout << endl;
      cout << "Error: " << CStringConverter::TtoA(message) << endl;
      cout << endl;
      cout << "In command line: \"" << CStringConverter::TtoA(m_parser.GetParsedCommandLine()) << "\"" << endl;
      cout << endl;
   }

   cout << "Usage: " << serverName << " -port xxxx" << endl;
   cout << endl;
   cout << "Command line parameters:" << endl;

   //cout << "         1         2         3         4         5         6         7         8" << endl;
   //cout << "12345678901234567890123456789012345678901234567890123456789012345678901234567890" << endl;
   //cout << "                            |" << endl;

   cout << " r -port                    The port to listen on." << endl;
   cout << " o -server                  The server address to bind to, in either" << endl;
   cout << "                            IPv4 dotted IP or IPv6 hex formats." << endl;

   if (m_type & TCPClient)
   {
      cout << "                            Defaults to IPv4 localhost." << endl;
   }
   else
   {
      cout << "                            Defaults to IPv4 INADDR_ANY." << endl;
   }

   cout << " o -numberOfIOThreads       Defaults to 0 (2 x processors)" << endl;
   cout << " o -socketPoolSize          Defaults to 10." << endl;
   cout << " o -noBufferPool            Turn off buffer pooling. " << endl;
   cout << " o -bufferPoolSize          Defaults to 10." << endl;

   if (m_type & SupportsPageAllocator)
   {
      cout << " o -bufferSize              Defaults to 1024 bytes OR" << endl;
      cout << "                            1 page if -pageSizedBuffers is also specified." << endl;
   }
   else
   {
      cout << " o -bufferSize              Defaults to 1024 bytes" << endl;
   }

   if ((m_type & TCPServer) || (m_type & UDPServer))
   {
      cout << " o -listenBacklog           Defaults to 5." << endl;
      cout << " o -maxConnections          Defaults to no limit" << endl;
   }

   if (m_type & TCPServer)
   {
      if ((m_type & DoesNotSupportWelcomeMessage) == 0)
      {
         cout << " o -noWelcomeMessage        Does not display a welcome message" << endl;
      }
   }

   if (m_type & SupportsPerfCounters)
   {
      cout << " o -registerCounters        Registers performance counters." << endl;
      cout << " o -unRegisterCounters      Removed performance counters." << endl;
   }

   if ((m_type & UDPMulticastServer) == UDPMulticastServer)
   {
      cout << " o -multicastAddress        Join the specified multicast group" << endl;
   }

   if (m_type & SupportsSendRecvBufSize)
   {
      cout << " o -recvBuffer              Allows you to set the socket receive buffer size." << endl;
      cout << "                            This can affect the TCP Receive Window size OR" << endl;
      cout << "                            allow the stack to buffer more UDP packets." << endl;
      cout << " o -sendBuffer              Allows you to set the send buffer size." << endl;
   }

   if (m_type & SupportsZeroByteRead)
   {
      cout << " o -syncZero                Uses synchronous zero byte reads." << endl;
      cout << " o -asyncZero               Uses asynchronous zero byte reads." << endl;
   }

   if (m_type & SupportsPageAllocator)
   {
      cout << " o -pageAlignedBuffers      Allocates buffers with page alignment." << endl;
      cout << " o -pageSizedBuffers        Causes -bufferSize to be treated as a value in." << endl;
      cout << "                            pages rather than in bytes." << endl;
   }

   if ((m_type & UDPConnectionServer) == UDPConnectionServer)
   {
      cout << " o -connectionTimeout       Aborts connections if there is no activity" << endl;
      cout << "                            on them for Y milliseconds." << endl;
      cout << "                            Defaults to 10000ms." << endl;
   }


   DEBUG_ONLY(cout << " o -displayDebug" << endl;)

   cout << " o -eventNamePrefix         Prefix shutdown and pause events with" << endl;
   cout << "                            the supplied name rather than \"JetByteTools\"" << endl;

   if ((m_type & SupportsFlowControl) == SupportsFlowControl)
   {
      cout << " o -maxPendingWrites        Flow control: the maximum number of writes" << endl;
      cout << "                            that can be outstanding before buffers are queued." << endl;
      cout << " o -maxBuffers              Flow control: The maximum number of buffers" << endl;
      cout << "                            that can be queued at one time." << endl;
      cout << " o -notify                  Flow control: The % change in queue usage between" << endl;
      cout << "                            notifications." << endl;
   }

   if ((m_type & SupportsProcessingDelay) == SupportsProcessingDelay)
   {
      cout << " o -delay                   The delay, in milliseconds, after each echo, before" << endl;
      cout << "                            the next read is issued. Defaults to 0ms." << endl;
   }

   if ((m_type & SupportsIOPoolEx) == SupportsIOPoolEx)
   {
      cout << " o -numEntries              The number of completions to retrieve with each" << endl;
      cout << "                            call to GetQueuedCompletionStatusEx(), defaults to 1." << endl;
   }

   cout << " o -perSocketBufferPoolSize The number of I/O buffers pooled in the socket." << endl;
   cout << "                            Defaults to 10." << endl;
   cout << " o -maxOps                  The maximum number of consecutive operations" << endl;
   cout << "                            per socket. Defaults to 10." << endl;

   ShowMoreHelp();

   cout << endl;
   cout << " r = required, o = optional" << endl;
   cout << endl;
   cout << "See http://www.lenholgate.com for more details." << endl;
}

void CCommandLine::ShowMoreHelp()
{

}

CCommandLine::Count CCommandLine::StringToCount(
   const _tstring &stringValue)
{
   const size_t value = _ttol(stringValue.c_str());

   if (value > static_cast<size_t>(std::numeric_limits<Count>::max()))
   {
      throw CException(_T("CCommandLine::StringToCount()"), _T("Value too big"));
   }

   return static_cast<Count>(value);
}

unsigned long CCommandLine::StringToUnsignedLong(
   const _tstring &stringValue)
{
   const size_t value = _ttol(stringValue.c_str());

   if (value > static_cast<size_t>(std::numeric_limits<unsigned long>::max()))
   {
      throw CException(_T("CCommandLine::StringToUnsignedLong()"), _T("Value too big"));
   }

   return static_cast<unsigned long>(value);
}
///////////////////////////////////////////////////////////////////////////////
// End of file: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
