///////////////////////////////////////////////////////////////////////////////
// File: SimpleServerShutdownHandler.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SimpleServerShutdownHandler.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/ToString.h"

#include "JetByteTools/Win32Tools/GlobalName.h"

#include "JetByteTools/SocketTools/IServerControl.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::CException;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;

using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;
using JetByteTools::Win32::CManualResetEvent;
using JetByteTools::Win32::CGlobalName;

using JetByteTools::Socket::IServerControl;

///////////////////////////////////////////////////////////////////////////////
// CSimpleServerShutdownHandler
///////////////////////////////////////////////////////////////////////////////

CSimpleServerShutdownHandler::CSimpleServerShutdownHandler()
   :  m_pServerControl(nullptr),
      m_sa(m_sd),
      m_pServerStartedEvent(nullptr),
      m_pServerStoppedEvent(nullptr)
{

}

CSimpleServerShutdownHandler::CSimpleServerShutdownHandler(
   IServerControl &serverControl)
   :  m_pServerControl(nullptr),
      m_sa(m_sd),
      m_pServerStartedEvent(nullptr),
      m_pServerStoppedEvent(nullptr)
{
   SetServerControl(serverControl);
}

CSimpleServerShutdownHandler::CSimpleServerShutdownHandler(
   IServerControl &serverControl,
   const _tstring &eventNamePrefix)
   :  m_pServerControl(nullptr),
      m_sa(m_sd),
      m_pServerStartedEvent(nullptr),
      m_pServerStoppedEvent(nullptr)
{
   SetServerControl(serverControl, eventNamePrefix);
}

CSimpleServerShutdownHandler::~CSimpleServerShutdownHandler()
{
   if (m_pServerStoppedEvent)
   {
      m_pServerStoppedEvent->Set();
   }

   delete m_pServerStartedEvent;
   delete m_pServerStoppedEvent;
}

void CSimpleServerShutdownHandler::SetServerControl(
   IServerControl &serverControl)
{
   SetServerControl(serverControl, _T("JetByteTools"));
}

void CSimpleServerShutdownHandler::SetServerControl(
   IServerControl &serverControl,
   const _tstring &eventNamePrefix)
{
   if (m_pServerControl)
   {
      throw CException(
            _T("CSimpleServerShutdownHandler::SetServerControl()"),
            _T("Server control is already set"));
   }

   m_pServerControl = &serverControl;

   m_pServerStartedEvent = new CManualResetEvent(CGlobalName(eventNamePrefix + _T("ServerStarted")), &m_sa);

   m_pServerStoppedEvent = new CManualResetEvent(CGlobalName(eventNamePrefix + _T("ServerStopped")), &m_sa);

   m_eventNamePrefix = eventNamePrefix;

   m_pServerStartedEvent->Set();
}

void CSimpleServerShutdownHandler::ShutdownNow()
{
   CManualResetEvent shutdownEvent(CGlobalName(m_eventNamePrefix + _T("ServerShutdown")), &m_sa);

   shutdownEvent.Set();
}

void CSimpleServerShutdownHandler::WaitForShutdownRequest()
{
   if (!m_pServerControl)
   {
      throw CException(
            _T("CSimpleServerShutdownHandler::WaitForShutdownRequest()"),
            _T("Server control is NOT set"));
   }

   CManualResetEvent shutdownEvent(CGlobalName(m_eventNamePrefix + _T("ServerShutdown")), &m_sa);
   CManualResetEvent pauseResumeEvent(CGlobalName(m_eventNamePrefix + _T("ServerPauseResume")), &m_sa);

   HANDLE handlesToWaitFor[2];

   handlesToWaitFor[0] = shutdownEvent.GetWaitHandle();
   handlesToWaitFor[1] = pauseResumeEvent.GetWaitHandle();

   bool accepting = true;
   bool done = false;

   while (!done)
   {
      DWORD waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, INFINITE);

      if (waitResult == WAIT_OBJECT_0)
      {
         done = true;
      }
      else if (waitResult == WAIT_OBJECT_0 + 1)
      {
         if (accepting)
         {
            m_pServerControl->StopAcceptingConnections();
         }
         else
         {
            m_pServerControl->StartAcceptingConnections();
         }

         accepting = !accepting;
      }
      else
      {
         throw CException(
            _T("CSimpleServerShutdownHandler::WaitForShutdownRequest()"),
            _T("Unexpected result from WaitForMultipleObjects - ") + ToString(waitResult));
      }
   }

   m_pServerStartedEvent->Reset();
}

///////////////////////////////////////////////////////////////////////////////
// End of file: CSimpleServerShutdownHandler.cpp
///////////////////////////////////////////////////////////////////////////////
